#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <string.h>
#include <unistd.h>

#define VWIDTH 640
#define VHEIGHT 480
#define CHARWIDTH 16
#define CHARHEIGHT 16
#define FONT_CHARS 6889
#define BRACK 0
#define WHITE 15

unsigned char *vbuf;
int fontcode_table[65565];

int clear_screen()
{
    int i;

    for (i=0; i<=VWIDTH * VHEIGHT; i++){
        vbuf[i] = 0;
    }
    return 0;
}

void disp(void)
{
    int pageoffset = 0;

    vga_setdisplaystart(0);
    vga_copytoplanar16(vbuf, WIDTH,
                       pageoffset, 80, VWIDTH, VHEIGHT);
    vga_setdisplaystart(pageoffset);
}

int JIS208(u_char ch1, u_char ch2)
{
    if (ch1 > 0x2A)
	return((ch2 - 0x41 + (ch1 - 0x26) * 96) << 5);
    else
	return((ch2 - 0x21 + (ch1 - 0x21) * 96) << 5);
}

int ccprint(int code, int x, int y, unsigned char *fontbase)
{
    char c_buf[CHARWIDTH];
    int i, j;
    int charbase;

    if ( x<0 || x>(VWIDTH-CHARWIDTH) || y<0 || y>(VHEIGHT-CHARHEIGHT) ){
        perror("irregal font location");
        return 0;
    }
    charbase = code * CHARWIDTH/4 * CHARHEIGHT;
    for (i=0; i<CHARHEIGHT; i++){
        for (j=0; j<CHARWIDTH/4; j++){
            c_buf[j*4]   = (fontbase[charbase + i*CHARWIDTH/4 + j]>>3) & 0x01;
            c_buf[j*4+1] = (fontbase[charbase + i*CHARWIDTH/4 + j]>>2) & 0x01;
            c_buf[j*4+2] = (fontbase[charbase + i*CHARWIDTH/4 + j]>>1) & 0x01;
            c_buf[j*4+3] = (fontbase[charbase + i*CHARWIDTH/4 + j])    & 0x01;
        }
        for (j=0; j<CHARWIDTH; j++){
           vbuf[(y+i)*VWIDTH+(x+j)] = c_buf[j] * WHITE;
        }
    }
    return 0;
}

int fontload(char *fontname, unsigned char *fontbase)
{
    FILE *fp;
    char buf[128];
    char c_buf[CHARWIDTH/4];
    int fontnum=0;
    int charbase = 0;
    int encode_num;
    int i,j;

    for (i=0; i<=65565; i++){ fontcode_table[i] = 0; }
    fp=fopen(fontname,"r");
    if (fp==NULL){
        perror("no 16x16 dot bdf font file\n");
        exit (0);
    }
    while (! fgets(buf, 128, fp)==NULL){
	if (!strncmp("ENCODING", buf, 8)) {
	    sscanf(buf, "ENCODING %d", &encode_num);
            fontcode_table[encode_num] = fontnum;
/*	    printf("%x\n", encode_num); */
	}
	if (!strncmp("BITMAP", buf, 6)) {
            for (i=0; i<CHARHEIGHT; i++){
                fgets(buf, 128, fp);
                for (j=0; j<(CHARWIDTH/4); j++){
                    if (buf[j] >='0' && buf[j] <='9') c_buf[j]=buf[j]-'0';
                    if (buf[j] >='a' && buf[j] <='f') c_buf[j]=buf[j]-'a'+10;
                    if (buf[j] >='A' && buf[j] <='F') c_buf[j]=buf[j]-'A'+10;
                    charbase = fontnum * CHARWIDTH/4 * CHARHEIGHT;
                    fontbase[ charbase + i*CHARWIDTH/4 + j] = c_buf[j];
                };
            }
            fontnum++;
        }
    }
    fclose (fp);
    return 0;
}

int main(int argc, char **argv)
{
    char *fontname, *filename;
    unsigned char *fontbase;
    unsigned char buf[80];
    FILE *fp;
    int ch;
    int i;
    int cursol_x, cursol_y, cursol_buf, kanji_mode=0;

    fontname = argv[1];
    filename = argv[2];
    fontbase = malloc(CHARWIDTH * CHARHEIGHT * FONT_CHARS);
    fontload(fontname, fontbase);
    vbuf = malloc(VWIDTH * VHEIGHT);
    vga_init();
    gl_setcontextvirtual(VWIDTH, VHEIGHT, 1, 8, vbuf);
    vga_setmode(G640x480x16);
    vga_clear();

    fp=fopen(filename,"r");
    if (fp==NULL){
        perror("no input file\n");
        exit (0);
    }

    cursol_y=0;
    while (! fgets(buf, 80, fp)==NULL){
        cursol_x=0;
        cursol_buf=0;
        while ( !(buf[cursol_buf]==0x0a) && cursol_buf<80){
            if (buf[cursol_buf]==0x1b){
                if (buf[cursol_buf+1]==0x28 && buf[cursol_buf+2]==0x42){
                    kanji_mode=0;
                }
                if (buf[cursol_buf+1]==0x24 && buf[cursol_buf+2]==0x42){
                    kanji_mode=1;
                }
                cursol_buf += 3;
            }
            if (kanji_mode==0){
                cursol_x += 8;
                cursol_buf += 1;
            }
            if (kanji_mode==1){
                ch = buf[cursol_buf] * 256 + buf[cursol_buf+1];
                ccprint(fontcode_table[ch], cursol_x, cursol_y, fontbase);
                cursol_x += 16;
                cursol_buf += 2;
            }
        }
        for (i=0; i<80; i++){buf[i]=0;}
        cursol_y += 16;
        if (cursol_y >= 480){
            disp();
            cursol_y = 0;
            for(;;){if (vga_getkey()) break;}
            clear_screen();
        }
    }
    disp();
    for(;;){if (vga_getkey()) break;}

    fclose (fp);
    vga_setmode(TEXT);
    return 0;
}
