#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <string.h>

#define MAXBUF 256
#define VWIDTH 640
#define VHEIGHT 480
#define CHARWIDTH 16
#define CHARHEIGHT 16
#define CCFONT_CHARS 6889
#define ACFONT_CHARS 191
#define BRACK 0
#define WHITE 15

unsigned char *vbuf;
int ccfontcode_table[65565];
int acfontcode_table[255];

int clear_screen()
{
    int i;

    for (i=0; i<=VWIDTH * VHEIGHT; i++){
        vbuf[i] = 0;
    }
    return 0;
}

void disp(void)
{
    int pageoffset = 0;

    vga_setdisplaystart(0);
    vga_copytoplanar16(vbuf, WIDTH,
                       pageoffset, 80, VWIDTH, VHEIGHT);
    vga_setdisplaystart(pageoffset);
}

int ccprint(int code, int x, int y, unsigned char *ccfontbase)
{
    char c_buf[CHARWIDTH];
    int i, j;
    int charbase;

    if ( x<0 || x>(VWIDTH-CHARWIDTH) || y<0 || y>(VHEIGHT-CHARHEIGHT) ){
        perror("irregal font location");
        return 0;
    }
    charbase = code * CHARWIDTH/4 * CHARHEIGHT;
    for (i=0; i<CHARHEIGHT; i++){
        for (j=0; j<CHARWIDTH/4; j++){
            c_buf[j*4]   = (ccfontbase[charbase + i*CHARWIDTH/4 + j]>>3) & 0x01;
            c_buf[j*4+1] = (ccfontbase[charbase + i*CHARWIDTH/4 + j]>>2) & 0x01;
            c_buf[j*4+2] = (ccfontbase[charbase + i*CHARWIDTH/4 + j]>>1) & 0x01;
            c_buf[j*4+3] = (ccfontbase[charbase + i*CHARWIDTH/4 + j])    & 0x01;
        }
        for (j=0; j<CHARWIDTH; j++){
           vbuf[(y+i)*VWIDTH+(x+j)] = c_buf[j] * WHITE;
        }
    }
    return 0;
}

int acprint(int code, int x, int y, unsigned char *acfontbase)
{
    char c_buf[CHARWIDTH/2];
    int i, j;
    int charbase;

    if ( x<0 || x>(VWIDTH-CHARWIDTH/2) || y<0 || y>(VHEIGHT-CHARHEIGHT) ){
        perror("irregal font location");
        return 0;
    }
    charbase = code * CHARWIDTH/8 * CHARHEIGHT;
    for (i=0; i<CHARHEIGHT; i++){
        for (j=0; j<CHARWIDTH/8; j++){
            c_buf[j*4]   = (acfontbase[charbase + i*CHARWIDTH/8 + j]>>3) & 0x01;
            c_buf[j*4+1] = (acfontbase[charbase + i*CHARWIDTH/8 + j]>>2) & 0x01;
            c_buf[j*4+2] = (acfontbase[charbase + i*CHARWIDTH/8 + j]>>1) & 0x01;
            c_buf[j*4+3] = (acfontbase[charbase + i*CHARWIDTH/8 + j])    & 0x01;
        }
        for (j=0; j<CHARWIDTH/2; j++){
           vbuf[(y+i)*VWIDTH+(x+j)] = c_buf[j] * WHITE;
        }
    }
    return 0;
}

int ccfontload(char *fontname, unsigned char *ccfontbase)
{
    FILE *fp;
    char buf[128];
    char c_buf[CHARWIDTH/4];
    int fontnum=0;
    int charbase = 0;
    int encode_num;
    int i,j;

    for (i=0; i<=65565; i++){ ccfontcode_table[i] = 0; }
    fp=fopen(fontname,"r");
    if (fp==NULL){
        perror("no 16x16 dot bdf font file\n");
        exit (0);
    }
    while (! fgets(buf, 128, fp)==NULL){
	if (!strncmp("ENCODING", buf, 8)) {
	    sscanf(buf, "ENCODING %d", &encode_num);
            ccfontcode_table[encode_num] = fontnum;
	}
	if (!strncmp("BITMAP", buf, 6)) {
            for (i=0; i<CHARHEIGHT; i++){
                fgets(buf, 128, fp);
                for (j=0; j<(CHARWIDTH/4); j++){
                    if (buf[j] >='0' && buf[j] <='9') c_buf[j]=buf[j]-'0';
                    if (buf[j] >='a' && buf[j] <='f') c_buf[j]=buf[j]-'a'+10;
                    if (buf[j] >='A' && buf[j] <='F') c_buf[j]=buf[j]-'A'+10;
                    charbase = fontnum * CHARWIDTH/4 * CHARHEIGHT;
                    ccfontbase[ charbase + i*CHARWIDTH/4 + j] = c_buf[j];
                };
            }
            fontnum++;
        }
    }
    fclose (fp);
    return 0;
}

int acfontload(char *fontname, unsigned char *acfontbase)
{
    FILE *fp;
    char buf[128];
    char c_buf[CHARWIDTH/8];
    int fontnum=0;
    int charbase = 0;
    int encode_num;
    int i,j;

    for (i=0; i<256; i++){ acfontcode_table[i] = 0; }
    fp=fopen(fontname,"r");
    if (fp==NULL){
        perror("no 8x16 dot bdf font file\n");
        exit (0);
    }
    while (! fgets(buf, 128, fp)==NULL){
	if (!strncmp("ENCODING", buf, 8)) {
	    sscanf(buf, "ENCODING %d", &encode_num);
            acfontcode_table[encode_num] = fontnum;
	}
	if (!strncmp("BITMAP", buf, 6)) {
            for (i=0; i<CHARHEIGHT; i++){
                fgets(buf, 128, fp);
                for (j=0; j<(CHARWIDTH/8); j++){
                    if (buf[j] >='0' && buf[j] <='9') c_buf[j]=buf[j]-'0';
                    if (buf[j] >='a' && buf[j] <='f') c_buf[j]=buf[j]-'a'+10;
                    if (buf[j] >='A' && buf[j] <='F') c_buf[j]=buf[j]-'A'+10;
                    charbase = fontnum * CHARWIDTH/8 * CHARHEIGHT;
                    acfontbase[ charbase + i*CHARWIDTH/8 + j] = c_buf[j];
                };
            }
            fontnum++;
        }
    }
    fclose (fp);
    return 0;
}

int main(int argc, char **argv)
{
    char *cfontname, *afontname, *filename;
    unsigned char *ccfontbase, *acfontbase;
    unsigned char buf[MAXBUF];
    FILE *fp;
    int ch;
    int i;
    int cursol_x, cursol_y, cursol_buf, kanji_mode=0;

    cfontname = argv[1];
    afontname = argv[2];
    filename = argv[3];
    ccfontbase = malloc(CHARWIDTH * CHARHEIGHT * CCFONT_CHARS);
    acfontbase = malloc(CHARWIDTH/2 * CHARHEIGHT * ACFONT_CHARS);
    ccfontload(cfontname, ccfontbase);
    acfontload(afontname, acfontbase);
    vbuf = malloc(VWIDTH * VHEIGHT);
    vga_init();
    gl_setcontextvirtual(VWIDTH, VHEIGHT, 1, 8, vbuf);
    vga_setmode(G640x480x16);
    vga_clear();

    fp=fopen(filename,"r");
    if (fp==NULL){
        perror("no input file\n");
        exit (0);
    }

    cursol_y=0;
    for (i=0; i<MAXBUF; i++){buf[i]=0;}
    while (! fgets(buf, MAXBUF, fp)==NULL){
        cursol_x=0;
        cursol_buf=0;
        while ( !(buf[cursol_buf]==0x0a) && cursol_buf<MAXBUF){
            if (buf[cursol_buf]==0x1b){
                if (buf[cursol_buf+1]==0x28 && buf[cursol_buf+2]==0x42){
                    kanji_mode=0;
                }
                if (buf[cursol_buf+1]==0x24 && buf[cursol_buf+2]==0x42){
                    kanji_mode=1;
                }
                cursol_buf += 3;
            }
            if (kanji_mode==0){
                ch = buf[cursol_buf];
                acprint(acfontcode_table[ch], cursol_x, cursol_y, acfontbase);
                cursol_x += 8;
                cursol_buf += 1;
            }
            if (kanji_mode==1){
                ch = buf[cursol_buf] * 256 + buf[cursol_buf+1];
                ccprint(ccfontcode_table[ch], cursol_x, cursol_y, ccfontbase);
                cursol_x += 16;
                cursol_buf += 2;
            }
        }
        for (i=0; i<MAXBUF; i++){buf[i]=0;}
        cursol_y += 16;
        if (cursol_y >= 480){
            disp();
            cursol_y = 0;
            for(;;){if (vga_getkey()) break;}
            clear_screen();
        }
    }
    disp();
    for(;;){if (vga_getkey()) break;}

    fclose (fp);
    vga_setmode(TEXT);
    return 0;
}
